#!/usr/bin/env python
import argparse
from beaupy import select  # type: ignore
from beaupy.spinners import Spinner  # type: ignore
from nyxfall.card import Card
from nyxfall.scryfall_requester import (
    search_exact,
    search_query,
    search_random,
)


def main():
    args = parse_args()
    run_cli(args)


def parse_args() -> argparse.Namespace:
    """Parses CLI arguments

    Returns:
        ``argparse.Namespace`` dictionary of arguments and their values
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "query", help="query to run against Scryfall", nargs="?"
    )
    parser.add_argument(
        "-e",
        "--exact",
        help="try and match the query with an exact card name",
        action="store_true",
    )
    parser.add_argument(
        "-r", "--random", help="fetch a random card", action="store_true"
    )
    parser.add_argument(
        "-a",
        "--ascii",
        help="renders the card frame using only basic ASCII characters",
        action="store_true",
    )
    return parser.parse_args()


def run_cli(args: argparse.Namespace):
    """Runs the CLI application based on the arguments provided"""
    if not args.query and not args.random:
        print("You must either supply a query or use the --random flag")
    elif args.random:
        for face in search_random().faces:
            print(face.format_as_card(ascii_only=args.ascii))
    elif args.exact:
        card = search_exact(args.query)
        if card is not None:
            for face in card.faces:
                print(face.format_as_card(ascii_only=args.ascii))
        else:
            print(f"Could not find a card with the name '{args.query}'")
    else:
        spinner = Spinner(text="Fetching cards")
        spinner.start()
        cards = search_query(args.query)
        spinner.stop()
        if len(cards) == 1:
            for face in cards[0].faces:
                print(face.format_as_card(ascii_only=args.ascii))
        elif cards:
            selected_card: Card = select(
                options=cards,  # type: ignore
                preprocessor=lambda card: card.name,
                pagination=True,
                page_size=7,
            )
            for face in selected_card.faces:
                print(face.format_as_card(ascii_only=args.ascii))
        else:
            print(f"Could not find any cards matchng the query '{args.query}'")


if __name__ == "__main__":
    main()



from nyxfall.card_face import CardFace


def test_extended_ascii_no_pt():
    expected_str = (
        "\n".join(
            [
                "┌──────────────────────────────────┐",
                "│┌────────────────────────────────┐│",
                "││Lightning Bolt               {R}││",
                "│└┬──────────────────────────────┬┘│",
                "│ │                              │ │",
                "│┌┴──────────────────────────────┴┐│",
                "││Instant                         ││",
                "│└┬──────────────────────────────┬┘│",
                "│ │Lightning Bolt deals 3 damage │ │",
                "│ │to any target.                │ │",
                "│ │ ──────────────────────────── │ │",
                "\x1b[3m│ │The sparkmage shrieked,       │ │",
                "│ │calling on the rage of the    │ │",
                "│ │storms of his youth. To his   │ │",
                "│ │surprise, the sky responded   │ │",
                "│ │with a fierce energy he'd     │ │",
                "│ │never thought to see again.   │ │\x1b[23m",
                "│ └──────────────────────────────┘ │",
                "│ CLU                              │",
                "└──────────────────────────────────┘",
            ]
        )
        + "\n"
    )
    card = CardFace(
        name="Lightning Bolt",
        scryfall_uri="",
        mana_cost="{R}",
        type_line="Instant",
        power=None,
        toughness=None,
        oracle_text="Lightning Bolt deals 3 damage to any target.",
        flavor_text="The sparkmage shrieked, calling on the rage of the storms of his youth. To his surprise, the sky responded with a fierce energy he'd never thought to see again.",
        set="CLU",
    )

    assert card.format_as_card() == expected_str


def test_extended_ascii_pt():
    expected_str = (
        "\n".join(
            [
                "┌──────────────────────────────────┐",
                "│┌────────────────────────────────┐│",
                "││Colossal Dreadmaw      {4}{G}{G}││",
                "│└┬──────────────────────────────┬┘│",
                "│ │                              │ │",
                "│┌┴──────────────────────────────┴┐│",
                "││Creature — Dinosaur             ││",
                "│└┬──────────────────────────────┬┘│",
                "│ │Trample (This creature can    │ │",
                "│ │deal excess combat damage to  │ │",
                "│ │the player or planeswalker    │ │",
                "│ │it's attacking.)              │ │",
                "│ │ ──────────────────────────── │ │",
                "\x1b[3m│ │If you feel the ground quake, │ │",
                "│ │run. If you hear its bellow,  │ │",
                "│ │flee. If you see its teeth,   │ │",
                "│ │it's too late.                │ │\x1b[23m",
                "│ │                       ┌─────┐│ │",
                "│ └───────────────────────┤ 6/6 ├┘ │",
                "│ M21                     └─────┘  │",
                "└──────────────────────────────────┘",
            ]
        )
        + "\n"
    )
    card = CardFace(
        name="Colossal Dreadmaw",
        scryfall_uri="",
        mana_cost="{4}{G}{G}",
        type_line="Creature — Dinosaur",
        power="6",
        toughness="6",
        oracle_text="Trample (This creature can deal excess combat damage to the player or planeswalker it's attacking.)",
        flavor_text="If you feel the ground quake, run. If you hear its bellow, flee. If you see its teeth, it's too late.",
        set="M21",
    )

    assert card.format_as_card() == expected_str


def test_basic_ascii_no_pt():
    expected_str = (
        "\n".join(
            [
                "+----------------------------------+",
                "|+--------------------------------+|",
                "||Lightning Bolt               {R}||",
                "|++------------------------------++|",
                "| |                              | |",
                "|++------------------------------++|",
                "||Instant                         ||",
                "|++------------------------------++|",
                "| |Lightning Bolt deals 3 damage | |",
                "| |to any target.                | |",
                "| | ---------------------------- | |",
                "| |The sparkmage shrieked,       | |",
                "| |calling on the rage of the    | |",
                "| |storms of his youth. To his   | |",
                "| |surprise, the sky responded   | |",
                "| |with a fierce energy he'd     | |",
                "| |never thought to see again.   | |",
                "| +------------------------------+ |",
                "| CLU                              |",
                "+----------------------------------+",
            ]
        )
        + "\n"
    )

    card = CardFace(
        name="Lightning Bolt",
        scryfall_uri="",
        mana_cost="{R}",
        type_line="Instant",
        power=None,
        toughness=None,
        oracle_text="Lightning Bolt deals 3 damage to any target.",
        flavor_text="The sparkmage shrieked, calling on the rage of the storms of his youth. To his surprise, the sky responded with a fierce energy he'd never thought to see again.",
        set="CLU",
    )

    assert card.format_as_card(ascii_only=True) == expected_str


def test_basic_ascii_pt():
    expected_str = (
        "\n".join(
            [
                "+----------------------------------+",
                "|+--------------------------------+|",
                "||Colossal Dreadmaw      {4}{G}{G}||",
                "|++------------------------------++|",
                "| |                              | |",
                "|++------------------------------++|",
                "||Creature — Dinosaur             ||",
                "|++------------------------------++|",
                "| |Trample (This creature can    | |",
                "| |deal excess combat damage to  | |",
                "| |the player or planeswalker    | |",
                "| |it's attacking.)              | |",
                "| | ---------------------------- | |",
                "| |If you feel the ground quake, | |",
                "| |run. If you hear its bellow,  | |",
                "| |flee. If you see its teeth,   | |",
                "| |it's too late.                | |",
                "| |                       +-----+| |",
                "| +-----------------------| 6/6 |+ |",
                "| M21                     +-----+  |",
                "+----------------------------------+",
            ]
        )
        + "\n"
    )

    card = CardFace(
        name="Colossal Dreadmaw",
        scryfall_uri="",
        mana_cost="{4}{G}{G}",
        type_line="Creature — Dinosaur",
        power="6",
        toughness="6",
        oracle_text="Trample (This creature can deal excess combat damage to the player or planeswalker it's attacking.)",
        flavor_text="If you feel the ground quake, run. If you hear its bellow, flee. If you see its teeth, it's too late.",
        set="M21",
    )

    assert card.format_as_card(ascii_only=True) == expected_str


def test_long_name_no_pt():
    expected_str = (
        "\n".join(
            [
                "┌──────────────────────────────────────┐",
                "│┌────────────────────────────────────┐│",
                "││Realmbreaker, the Invasion Tree  {3}││",
                "│└┬──────────────────────────────────┬┘│",
                "│ │                                  │ │",
                "│┌┴──────────────────────────────────┴┐│",
                "││Legendary Artifact                  ││",
                "│└┬──────────────────────────────────┬┘│",
                "│ │{2}, {T}: Target opponent mills   │ │",
                "│ │three cards. Put a land card from │ │",
                "│ │their graveyard onto the          │ │",
                "│ │battlefield tapped under your     │ │",
                '│ │control. It gains "If this land   │ │',
                "│ │would leave the battlefield, exile│ │",
                "│ │it instead of putting it anywhere │ │",
                '│ │else."                            │ │',
                "│ │{10}, {T}, Sacrifice Realmbreaker:│ │",
                "│ │Search your library for any number│ │",
                "│ │of Praetor cards, put them onto   │ │",
                "│ │the battlefield, then shuffle.    │ │",
                "│ └──────────────────────────────────┘ │",
                "│ MOM                                  │",
                "└──────────────────────────────────────┘",
            ]
        )
        + "\n"
    )

    card = CardFace(
        name="Realmbreaker, the Invasion Tree",
        scryfall_uri="",
        mana_cost="{3}",
        type_line="Legendary Artifact",
        power=None,
        toughness=None,
        oracle_text='{2}, {T}: Target opponent mills three cards. Put a land card from their graveyard onto the battlefield tapped under your control. It gains "If this land would leave the battlefield, exile it instead of putting it anywhere else."\n{10}, {T}, Sacrifice Realmbreaker: Search your library for any number of Praetor cards, put them onto the battlefield, then shuffle.',
        flavor_text=None,
        set="MOM",
    )

    assert card.format_as_card() == expected_str


def test_long_type_line_pt():
    expected_str = (
        "\n".join(
            [
                "┌───────────────────────────────────────────────────┐",
                "│┌─────────────────────────────────────────────────┐│",
                "││The Reality Chip                           {1}{U}││",
                "│└┬───────────────────────────────────────────────┬┘│",
                "│ │                                               │ │",
                "│┌┴───────────────────────────────────────────────┴┐│",
                "││Legendary Artifact Creature - Equipment Jellyfish││",
                "│└┬───────────────────────────────────────────────┬┘│",
                "│ │You may look at the top card of your library   │ │",
                "│ │any time.                                      │ │",
                "│ │As long as The Reality Chip is attached to a   │ │",
                "│ │creature, you may play lands and cast spells   │ │",
                "│ │from the top of your library.                  │ │",
                "│ │Reconfigure {2}{U} ({2}{U}: Attach to target   │ │",
                "│ │creature you control; or unattach from a       │ │",
                "│ │creature. Reconfigure only as a sorcery. While │ │",
                "│ │attached, this isn't a creature.)              │ │",
                "│ │                                        ┌─────┐│ │",
                "│ └────────────────────────────────────────┤ 0/4 ├┘ │",
                "│ NEO                                      └─────┘  │",
                "└───────────────────────────────────────────────────┘",
            ]
        )
        + "\n"
    )

    card = CardFace(
        name="The Reality Chip",
        scryfall_uri="",
        mana_cost="{1}{U}",
        type_line="Legendary Artifact Creature - Equipment Jellyfish",
        power="0",
        toughness="4",
        oracle_text="You may look at the top card of your library any time.\nAs long as The Reality Chip is attached to a creature, you may play lands and cast spells from the top of your library.\nReconfigure {2}{U} ({2}{U}: Attach to target creature you control; or unattach from a creature. Reconfigure only as a sorcery. While attached, this isn't a creature.)",
        flavor_text=None,
        set="NEO",
    )

    assert card.format_as_card() == expected_str



import time
import requests
from typing import Any, Optional
from nyxfall.card import Card
from nyxfall.card_face import CardFace

SCRYFALL_BASE = "https://api.scryfall.com/cards/"
HEADERS = {"User-Agent": "NyxfallApp/0.0.1", "Accept": "*/*"}


def search_exact(name: str) -> Optional[Card]:
    """Searches for a card with the name exactly matching a string

    Args:
        name: Name of card to match

    Returns:
        ``Card`` object matching that string if one was found, None otherwise
    """
    req = requests.get(f"{SCRYFALL_BASE}named?exact={name}", headers=HEADERS)
    if req.status_code != requests.codes.ok:
        return None
    return _map_response(req.json())


def search_random() -> Card:
    """Searches for a random card

    Returns:
        ``Card`` object of a random card
    """
    return _map_response(
        requests.get(f"{SCRYFALL_BASE}random", headers=HEADERS).json()
    )


def search_query(query: str) -> list[Card]:
    """Searches for a query and returns all cards that match

    Args:
        query: Query to execute

    Returns:
        All ``Card`` objects matching the query, or an empty list of no cards were found
    """
    response = requests.get(
        f"{SCRYFALL_BASE}search?q={query}+game:paper&page=1", headers=HEADERS
    ).json()
    card_data = [_map_response(card) for card in response.get("data", [])]

    # Traverse pagination from responses
    while response.get("has_more", False):
        # Rate limit ourselves by 100ms between requests
        time.sleep(100 / 1000)
        response = requests.get(response.get("next_page", "")).json()
        card_data += [_map_response(card) for card in response.get("data", [])]

    return card_data


def _map_response(response: dict[str, Any]) -> Card:
    if response.get("card_faces") is None:
        return Card(
            faces=[_map_card_face(response)], name=response.get("name", "")
        )

    return Card(
        faces=[
            _map_card_face(face) for face in response.get("card_faces", [])
        ],
        name=response.get("name", ""),
    )


def _map_card_face(face: dict[str, Any]) -> CardFace:
    return CardFace(
        name=face.get("name", ""),
        scryfall_uri=face.get("scryfall_uri", ""),
        mana_cost=face.get("mana_cost", ""),
        type_line=face.get("type_line", ""),
        power=face.get("power", None),
        toughness=face.get("toughness", None),
        oracle_text=face.get("oracle_text", ""),
        flavor_text=face.get("flavor_text", None),
        set=face.get("set", "").upper(),
    )



from dataclasses import dataclass
from typing import Optional
from textwrap import fill

CARD_TEXT_DEFAULT_WIDTH = 32
# When rendering a card, leave a reasonable space between the end of the name and the start of the mana cost
NAME_MANA_COST_GAP = 2


@dataclass
class CardFace:
    """Data required to display an MTG card

    Attributes:
        name: Colossal Dreadmaw
        scryfall_uri: https://scryfall.com/card/m21/176/colossal-dreadmaw
        mana_cost: {4}{G}{G}
        type_line: Creature — Dinosaur
        power: 6
        toughness: 6
        oracle_text: Trample (This creature can deal excess combat damage to the player or planeswalker it's attacking.)
        flavor_text: If you feel the ground quake, run. If you hear its bellow, flee. If you see its teeth, it's too late.
        set: XLN
    """

    name: str
    scryfall_uri: str
    mana_cost: str
    type_line: str
    power: Optional[str]
    toughness: Optional[str]
    oracle_text: str
    flavor_text: Optional[str]
    set: str

    def format_as_card(self, ascii_only: bool = False) -> str:
        """Builds a string that will display a ``Card`` similar to an actual MTG card

        Args:
            ascii_only: True if card frame should be rendered using only the basic ASCII set

        Returns:
            Formatted and newline-separated string that should display a card when printed
        """
        down_right = "+" if ascii_only else "┌"
        down_left = "+" if ascii_only else "┐"
        up_right = "+" if ascii_only else "└"
        up_left = "+" if ascii_only else "┘"
        vertical = "|" if ascii_only else "│"
        horizontal = "-" if ascii_only else "─"
        down_horizontal = "+" if ascii_only else "┬"
        up_horizontal = "+" if ascii_only else "┴"
        left_vertical = "|" if ascii_only else "┤"
        right_vertical = "|" if ascii_only else "├"

        # Defalt card width to 30 characters unless the card has a particularly long name or mana cost
        card_text_width = (
            CARD_TEXT_DEFAULT_WIDTH
            if max(
                (len(self.name) + len(self.mana_cost) + NAME_MANA_COST_GAP),
                len(self.type_line),
            )
            <= CARD_TEXT_DEFAULT_WIDTH
            else max(
                (len(self.name) + len(self.mana_cost) + NAME_MANA_COST_GAP),
                len(self.type_line),
            )
        )

        # Initialise a string list with everything up until the first conditonal section
        # of the display. It will be appended to later and joined with a newline separator at the end
        card = [
            # Top of outside bounding box
            f"{down_right}{horizontal * (card_text_width + 2)}{down_left}",
            # Name and mana cost
            f"{vertical}{down_right}{horizontal * card_text_width}{down_left}{vertical}",
            f"{vertical}{vertical}{self.name}{" " * (card_text_width - len(self.name) - len(self.mana_cost))}{self.mana_cost}{vertical}{vertical}",
            f"{vertical}{up_right}{down_horizontal}{horizontal * (card_text_width - 2)}{down_horizontal}{up_left}{vertical}",
            # Empty image box
            "\n".join(
                [
                    f"{vertical} {vertical}{" " * (card_text_width - 2)}{vertical} {vertical}"
                ]
                * 1
            ),
            # Type line
            f"{vertical}{down_right}{up_horizontal}{horizontal * (card_text_width - 2)}{up_horizontal}{down_left}{vertical}",
            f"{vertical}{vertical}{self.type_line}{" " * (card_text_width - len(self.type_line))}{vertical}{vertical}",
            f"{vertical}{up_right}{down_horizontal}{horizontal * (card_text_width - 2)}{down_horizontal}{up_left}{vertical}",
            # Oracle text
            self._wrap_and_pad(self.oracle_text, card_text_width, vertical),
        ]

        # Flavour text
        if self.flavor_text:
            card.append(
                f"{vertical} {vertical} {horizontal * (card_text_width - 4)} {vertical} {vertical}"
            )
            # Append and prepend flavour text with ANSI escape code for italics
            card.append(
                ("" if ascii_only else "\x1b[3m")
                + self._wrap_and_pad(
                    self.flavor_text, card_text_width, vertical
                )
                + ("" if ascii_only else "\x1b[23m")
            )

        if self.power and self.toughness:
            # Width of characters in power and toughness plus 3 for the forward slash and spacing
            pt_box_width = len(str(self.power)) + len(str(self.toughness)) + 3
            pt_box = [
                f"{vertical} {vertical}{" " * (card_text_width - pt_box_width - 4)}{down_right}{horizontal * pt_box_width}{down_left}{vertical} {vertical}",
                f"{vertical} {up_right}{horizontal * (card_text_width - pt_box_width - 4)}{left_vertical} {str(self.power)}/{str(self.toughness)} {right_vertical}{up_left} {vertical}",
                f"{vertical} {self.set}{" " * (card_text_width - len(self.set) - pt_box_width - 3)}{up_right}{horizontal * 5}{up_left}  {vertical}",
                f"{up_right}{horizontal * (card_text_width + 2)}{up_left}",
            ]
            card.extend(pt_box)
        else:
            set_box = [
                f"{vertical} {up_right}{horizontal * (card_text_width - 2)}{up_left} {vertical}",
                f"{vertical} {self.set}{" " * (card_text_width - len(self.set))} {vertical}",
                f"{up_right}{horizontal * (card_text_width + 2)}{up_left}",
            ]
            card.extend(set_box)

        return "\n".join(card) + "\n"

    def _wrap_and_pad(self, text: str, card_width: int, vert_char: str) -> str:
        """Helper function that breaks long strings on to newlines and pads each line

        Args:
            text: Text to be wrapped and padded (e.g. oracle text, flavour text)
            card_width: Width of the card (naturally the width text will be padded to)

        Returns:
            String of ``text`` with newline breaks and padded out to card width length
        """
        # Passing a string with newline characters to textwrap.fill() causes some ugly line
        # breaks, so we want to tell it to ignore linebreaks. This will create a list of strings
        # where each element is where we want to break, but each element will also include
        # the original newlines, so we need to split again on those newlines, flatten those
        # sub-lists and then finally return that list joined with newline as a separator
        return "\n".join(
            [
                f"{vert_char} {vert_char}{line}{" " * (card_width - len(line) - 2)}{vert_char} {vert_char}"
                for lines in [
                    fill(
                        paragraph,
                        width=card_width - 2,
                        replace_whitespace=False,
                    ).split("\n")
                    for paragraph in text.splitlines()
                ]
                for line in lines
            ]
        )



import argparse
from unittest.mock import patch
from pytest import CaptureFixture
from nyxfall.__main__ import run_cli


def test_cli_require_query_or_random(capfd: CaptureFixture[str]):
    args = argparse.Namespace(query=None, random=None)
    run_cli(args)
    assert (
        capfd.readouterr().out
        == "You must either supply a query or use the --random flag\n"
    )


def test_cli_search_random():
    with patch("nyxfall.__main__.search_random") as search_random:
        args = argparse.Namespace(query=None, random=True, ascii=False)
        run_cli(args)
        search_random.assert_called_once()


def test_cli_search_exact():
    with patch("nyxfall.__main__.search_exact") as search_exact:
        args = argparse.Namespace(
            query="Lightning Bolt", random=None, exact=True, ascii=False
        )
        run_cli(args)
        search_exact.assert_called_once_with("Lightning Bolt")
